use std::error::Error;
use super::util::xml_safe as x;
use crate::models::*;
use crate::templates::PAGE_SIZE;
use crate::util::*;
use std::{fs, io};
use std::fs::{create_dir_all, File};
use std::path::{PathBuf};
use mailparse::*;
use regex::Regex;
use std::fmt::Write as FmtWrite;
use std::io::Write as IoWrite;
use html5ever::tendril::TendrilSink;
use kuchiki::NodeRef;

fn header(title: &str, css_path: &str) -> String {
    format!(
        r##"<!DOCTYPE html>
        <html>
        <head>
        <title>{title}</title>
        <meta http-equiv='Permissions-Policy' content='interest-cohort=()'/>
        <link rel='stylesheet' type='text/css' href='{css_path}' />
        <meta name='viewport' content='width=device-width, initial-scale=1.0, maximum-scale=1.0,user-scalable=0' />
        <link rel='icon' href='data:image/svg+xml,<svg xmlns=%22http://www.w3.org/2000/svg%22 viewBox=%220 0 100 100%22><text y=%22.9em%22 font-size=%2290%22>📧</text></svg>'>
        <meta name="description" content="{title}"/>
        </head>
        <body>
        <div id="button-container">
            <!-- class 2 buttons will appear here -->
        </div>
            <button onclick="window.location.href='../../index.html'">Home</button>
             <input name="" type="button" value="Jump Item:">
             <input type="text" name="textfield" id="textfield" />
             <!--dropdown for options-->
            <div class="dropdown">
            <button>Options</button>
                <div class="dropdown-content">
                     <a href="#option1">Class 1</a>
                    <a href="#option2">Class 2</a>
                    <a href="#option3">Turbo Augment</a>
                </div>
            </div>
            <!-- Viewspecs button -->
        <button class="viewspecs-button" onclick="openNewWindow()">Viewspecs</button>
        </div>
    </div>

        <div>
            <b>sort by</b> [<a href='../../authors.html'>author</a>] [<a href='../../dates.html'>dates</a>] [<a href='../../subjects.html'>subjects</a>]
        </div>
        <hr>
        "##
    )
}

const FOOTER: &str = r#"
</body>
</html>
"#;

fn generate_popup_js() -> String {
    r#"
        function toggleNumbering() {
            const numberingCheckbox = document.getElementById('numbering');
            const purplenumbers = window.opener.document.querySelectorAll('.purplenumber');
            purplenumbers.forEach(element => {
                element.classList.toggle('transparent', !numberingCheckbox.checked);
            });
        }

        document.addEventListener('DOMContentLoaded', function() {
            const numberingCheckbox = document.getElementById('numbering');
            numberingCheckbox.addEventListener('change', toggleNumbering);
        });
    "#.to_string()
}


const JS_SCRIPT_VIEWSPECS: &str = r##"
<script>
    function showElement(body_id, summary_id){
        const summaries = document.querySelectorAll('.message-meta');
        const innerElements = document.querySelectorAll('.email-body');
        const zoom_ins = document.querySelectorAll('.zoom_in');
        const zoom_outs = document.querySelectorAll('.zoom_out');

        summaries.forEach(summary => {
            summary.style.display = 'none';
        });
        innerElements.forEach(element => {
            element.style.display = 'none';
        });
        
        zoom_ins.forEach(element => {
            element.style.display = 'none';
        });
        
        zoom_outs.forEach(element => {
            element.style.display = 'block';
        });

        document.getElementById(body_id).style.display = 'block';
        document.getElementById(summary_id).style.display = 'block';
        
        window.scrollTo(0, 0);
    }

    function showSummaries() {
        const summaries = document.querySelectorAll('.message-meta');
        const innerElements = document.querySelectorAll('.email-body');
        const zoom_ins = document.querySelectorAll('.zoom_in');
        const zoom_outs = document.querySelectorAll('.zoom_out');

        summaries.forEach(summary => {
            summary.style.display = 'block';
        });

        innerElements.forEach(element => {
            element.style.display = 'none';
        });
        
        zoom_ins.forEach(element => {
            element.style.display = 'block';
        });
        
        zoom_outs.forEach(element => {
            element.style.display = 'none';
        });
        
        window.scrollTo(0, 0);
    }

    function jumpToItem() {
            const id = document.getElementById('textfield').value;
            if (id) {
                window.location.href = `#${id}`;
                }
        }


        document.addEventListener('DOMContentLoaded', function() {
            const textfield = document.getElementById('textfield');
            textfield.addEventListener('keydown', function(event) {
                if (event.key === 'Enter') {
                    jumpToItem();
                }
            });

        const jumpButton = document.querySelector('input[type="button"][value="Jump Item:"]');
        jumpButton.addEventListener('click', jumpToItem);
        });

    function openNewWindow() {
        const popupJs = `
            function toggleNumbering() {
                const numberingCheckbox = document.getElementById('numbering');
                const purplenumbers = window.opener.document.querySelectorAll('.purplenumber');
                purplenumbers.forEach(element => {
                    element.classList.toggle('transparent', !numberingCheckbox.checked);
                });
            }

            document.addEventListener('DOMContentLoaded', function() {
                const numberingCheckbox = document.getElementById('numbering');
                numberingCheckbox.addEventListener('change', toggleNumbering);
            });
        `;

        const newWindow = window.open('', '', 'width=400,height=300');
        newWindow.document.write(`
            <html>
            <head>
                <title>Viewspecs</title>
                <style>
                    .popup-content {
                        font-family: Arial, sans-serif;
                        font-size: 14px;
                        margin: 20px;
                    }
                    .popup-content fieldset {
                        border: 1px solid #000;
                        padding: 10px;
                    }
                    .popup-content legend {
                        padding: 0 10px;
                        font-weight: bold;
                    }
                    .popup-content label {
                        display: block;
                        margin-bottom: 5px;
                    }
                    .popup-content input {
                        margin-right: 5px;
                    }
                    .accept-button {
                        margin-top: 10px;
                        padding: 5px 10px;
                        font-size: 14px;
                        cursor: pointer;
                    }
                    .transparent {
                        color: transparent !important;
                    }
                </style>
            </head>
            <body>
                <div class="popup-content">
                    <fieldset>
                        <legend>Show</legend>
                        <label><input type="checkbox" id="blankLines"> y z Blank lines</label>
                        <label><input type="checkbox" id="numbering" checked> m n Numbering</label>
                        <label><input type="checkbox" id="statementSignatures"> K L Statement signatures</label>
                    </fieldset>
                    <button class="accept-button" onclick="window.close()">Accept</button>
                </div>
                <script>${popupJs}<\/script>
            </body>
            </html>
        `);
        newWindow.document.close();
    }



</script>
"##;

impl Lists {
    pub fn to_html(&self) -> String {
        let mut lists: String = String::new();
        for list in &self.lists {
            lists.push_str(&format!(
                "<a href='./{0}'><h2>{0}</h2></a>\n",
                x(&list.config.name)
            ));
        }

        let body: String = format!(r#"<h1 class="page-title">Mail Archives</h1><hr>{lists}<hr>"#);
        header("Mail Archives", "style.css") + &body + FOOTER
    }
}

// currently 0-indexed
fn build_page_idx(current: usize, total: usize) -> String {
    let mut pages: String = String::new();
    if total == 1 {
        return pages;
    }

    for n in 0..total {
        if current == n {
            // writing to a string so discarding errors is fine
            let _ = write!(pages, "<b>{}</b>", n + 1);
        } else {
            let path = match n {
                0 => "index.html".to_string(),
                n => format!("index-{}.html", n + 1),
            };
            // writing to a string so discarding errors is fine
            let _ = write!(pages, r#"<a href="{}">{}</a>"#, path, n + 1);
        }
    }

    format!("<b>Pages</b>: {pages}<hr>")
}

impl List {
    pub fn to_html(&self) -> Vec<String> {
        let page_count = self.thread_topics.len() / PAGE_SIZE + 1;
        // hacky
        self.thread_topics
            .chunks(PAGE_SIZE)
            .enumerate()
            .map(|(n, thread_topics)| {
                let mut threads = String::new();
                for thread in thread_topics {
                    let _ = write!(
                        threads,
                        r#"
                        <div class='message-sum'>
                        <a class="bigger" href="threads/{path_id}.html">{subject}</a>
                        <br>
                        <div class="monospace">{preview}</div>
                        <b>{from}</b> | {replies} replies | {date}<hr>
                        "#,
                        path_id = thread.message.hash,
                        subject = x(&thread.message.subject),
                        replies = thread.reply_count,
                        // get only year-month-day part of datetime
                        date = thread.last_reply.to_rfc3339().split_once('T').unwrap().0,
                        from = x(thread
                            .message
                            .from
                            .name
                            .as_ref()
                            .unwrap_or(&thread.message.from.address)),
                        preview = x(&thread.message.preview),
                    );
                }
                let body = format!(
                    r#"<h1 class="page-title">
                    {title} <a href="atom.xml"><img alt="Atom feed" src='{rss_svg}' /></a>
                    </h1>
                    {description}<br>
                    <a href="mailto:{list_email}">{list_email}</a>
                    <hr>
                    {threads}
                    {page_idx}"#,
                    description = self.config.description,
                    title = self.config.title,
                    threads = threads,
                    page_idx = build_page_idx(n, page_count),
                    list_email = self.config.email,
                    rss_svg = RSS_SVG,
                );

                header(&self.config.title, "../style.css") + &body + FOOTER
            })
            .collect()
    }
}

impl MailAddress {
    fn to_html(&self) -> String {
        let mut out = String::new();
        if let Some(n) = &self.name {
            out.push('"');
            out.push_str(&x(n));
            out.push_str("\" ");
        }
        out.push_str(&format!(
            "<<a href='mailto:{0}'>{0}</a>>",
            &x(&self.address)
        ));
        out
    }
}

impl Thread {
    fn extract_html_content(&self, part: &ParsedMail) -> Option<String> {
        if part.ctype.mimetype == "text/html" {
            return part.get_body().ok();
        }
        for subpart in &part.subparts {
            if let Some(html) = self.extract_html_content(subpart) {
                return Some(html);
            }
        }
        None
    }

    fn retrieve_attachments(parsed_mail: ParsedMail, dir_parent: PathBuf, file_name: String) -> io::Result<Vec<PathBuf>> {
        // lists of all paths to attachments
        let mut paths = Vec::new();

        // iterating through each part of email
        for part in parsed_mail.subparts {
            let disposition = part.get_content_disposition();
            // if this is an attachment
            if disposition.disposition == DispositionType::Attachment {
                // creating directory for it
                let full_dir = dir_parent.clone().join(".attachments").join(file_name.clone());
                create_dir_all(full_dir.clone()).expect("Unable to create directory for attachments");

                let filename = disposition.params.get("filename")
                    .map(|s| s.to_string())
                    .unwrap_or_else(|| "unnamed_file".to_string());

                // creating full absolute path to the file
                let full_path = full_dir.join(filename.clone());
                let mut output = File::create(full_path.clone())?;

                // writing the data to the file
                output.write_all(&part.get_body_raw().unwrap())?;

                // retrieving relative path
                let relative_path: PathBuf = match full_path.strip_prefix(&dir_parent) {
                    Ok(relative_path) => relative_path.to_path_buf(),
                    Err(_) => full_path.clone(),
                };
                paths.push(relative_path);
            }
        }

        Ok(paths)
    }

    fn parse_email(&self, parsed_mail: &ParsedMail) -> Result<String, Box<dyn Error>> {
        if let Some(html_content) = self.extract_html_content(&parsed_mail) {
            return Ok(html_content.replace("\r\n", ""));
        } else if parsed_mail.ctype.mimetype.starts_with("text/plain") {
            // Convert plain text to simple HTML
            let plain_text = parsed_mail.get_body()?.replace("\r\n", "<br>");
            let html_content = format!("<html><body>{}</body></html>", plain_text);
            return Ok(html_content);
        }

        Err(From::from("No HTML or text content found"))
    }

    pub fn to_html(&self, out_dir: PathBuf) -> String {
        fn extract_styles(html: &str) -> String {
            let style_pattern = Regex::new(r"(?is)<style[^>]*>(.*?)</style>").unwrap();
            style_pattern.captures_iter(html)
                .filter_map(|caps| caps.get(1))
                .map(|m| m.as_str())
                .collect::<Vec<_>>()
                .join("\n")
        }

        fn purple_numbers(html: &str, link: &str)-> String {
            let process_html = |html: &str| -> String {
                let document = kuchiki::parse_html().one(html);
                let mut ref_number = 1;
                process_node(&document, &mut ref_number, &link);
                document.to_string()
            };

            fn process_node(node: &NodeRef, ref_number: &mut usize, link: &str) {
                // Create a vector of nodes to avoid borrowing issues while modifying the tree
                let children: Vec<NodeRef> = node.children().collect();
                for child in children {
                    if let Some(text_node) = child.as_text() {
                        if !text_node.borrow().trim().is_empty() {
                            let new_node = kuchiki::parse_html().one(format!(
                                r#"<span class="purplenumber" id="{}"><a href="{}html">{}</a></span>"#,
                                ref_number, link, ref_number,
                            ));
                            child.insert_after(new_node);
                            *ref_number += 1;
                        }
                    }
                    // Recursively process the children nodes
                    process_node(&child, ref_number, link);
                }
            }
            let result = process_html(html);
            result
        }



        fn remove_style_tags(html: &str) -> String {
            Regex::new(r"(?s)<style.*?>.*?</style>").unwrap()
                .replace_all(html, "")
                .to_string()
        }

        fn remove_inline_styles_from_body(html: &str) -> String {
            Regex::new(r"<body[^>]*>").unwrap()
                .replace(html, "<body>")
                .to_string()
        }

        fn extract_body(html: &str) -> String {
            let open_tag = "<body";
            let close_tag = "</body>";

            if let Some(start_pos) = html.find(open_tag) {
                if let Some(open_tag_end) = html[start_pos..].find('>') {
                    let body_start = start_pos + open_tag_end + 1;
                    if let Some(end_pos) = html[body_start..].find(close_tag) {
                        return html[body_start..body_start + end_pos].to_string();
                    }
                }
            }
            html.to_string()
        }

        let root = &self.messages[0];
        let mut body = header(&x(&root.subject), "../../style.css");

        let _ = write!(
            body,
            r#"<h1 class="page-title">{title}
        <a href="{path_id}.xml"><img alt="Atom Feed" src='{rss_svg}'></a>
        </h1>
        <div>
        <a href="../">Back</a>
        <hr>
        <div id='content'>"#,
            title = x(&root.subject),
            rss_svg = RSS_SVG,
            path_id = root.hash,
        );

        let mut message_id_css = 0;

        for msg in &self.messages {
            let in_reply_to = if let Some(irt) = &msg.in_reply_to {
                format!(
                    "In-Reply-To: <a href='#{0}'>{1}</a><br>\n",
                    x(irt),
                    x(&truncate_ellipsis(irt, 40))
                )
            } else {
                String::new()
            };

            let extra_headers = format!(
                "Message-Id: <a href='#{0}'>{0}</a><br>\nTo: {1}<br>\n{2}",
                x(&msg.id),
                msg.to.iter().map(|x| x.to_html()).collect::<Vec<String>>().join(","),
                if msg.cc.is_empty() {
                    String::new()
                } else {
                    format!("Cc: {}<br>\n", msg.cc.iter().map(|x| x.to_html()).collect::<Vec<String>>().join(","))
                }
            );

            let data = match fs::read_to_string(&msg.original_path) {
                Ok(content) => content,
                Err(e) => {
                    eprintln!("Error reading file: {}", e);
                    return "Error loading content".to_string();
                }
            };

            let parsed_mail = match parse_mail(&data.as_bytes()) {
                Ok(mail) => mail,
                Err(e) => return format!("Error parsing email: {}", e),
            };

            let html = self.parse_email(&parsed_mail).unwrap_or_else(|_| msg.body.clone());
            let styles = extract_styles(&html);
            let mut body_content = remove_style_tags(&extract_body(&html));
            body_content = purple_numbers(&*body_content, "#");
            
            let attachments_paths = Self::retrieve_attachments(parsed_mail, out_dir.clone(), msg.hash.clone())
                .expect("Unable to retrieve attachment");

            let attachments_list = attachments_paths.iter()
                .map(|path| format!("<li><a href='../{}'>{}</a></li>", path.to_string_lossy(), path.file_name().unwrap().to_string_lossy()))
                .collect::<String>();

            let attachments_html = if attachments_paths.is_empty() {
                String::new()
            } else {
                format!("<details><summary>Attachments</summary><ul>{}</ul></details>", attachments_list)
            };

            let _ = write!(
                body,
                r#"
            <div id="{msg_id}" class="message">
                <div id="summary-{message_id_css}" class="message-meta" style="position: relative;">
                    <div style="position: absolute; right: 100%; top: 0; height: 100%;">
                        <button class="zoom_in" onclick="showElement('email-body-{message_id_css}', 'summary-{message_id_css}')">In</button>
                        <button class="zoom_out" onclick="showSummaries()" style="display: none;">Out</button>
                    </div>
                    <span class="bold">
                        {subject}
                    </span>
                    <br>
                    From: {from}
                    <br>
                    Date: <span>{date}</span>
                    <br>
                    {in_reply_to}
                    <details>
                        <summary>More</summary>
                        {extra_headers}
                    </details>
                    <a class="bold" href="{mailto}">Reply</a>
                    [<a href="../messages/{msg_path}.mbox">Export</a>]
                    {attachments_html}
                </div>
                <div id="email-body-{message_id_css}" class="email-body" style="position: relative;">
                    <style>
                        #message_{message_id_css} {{
                            {styles}
                        }}
                    </style>
                    <div id="message_{message_id_css}">
                        {body_content}
                    </div>
                </div>
            </div>
            "#,
                msg_id = msg.hash,
                msg_path = msg.hash,
                subject = x(&msg.subject),
                mailto = x(&msg.mailto),
                from = msg.from.to_html(),
                date = msg.date,
                in_reply_to = in_reply_to,
                extra_headers = extra_headers,
                attachments_html = attachments_html,
                message_id_css = message_id_css,
                styles = styles,
                body_content = body_content,
            );



            
            // increment message id
            message_id_css += 1;
        }

        body.push_str(&*format!(r#"</div><hr>{}</body></html>"#, JS_SCRIPT_VIEWSPECS));
        remove_inline_styles_from_body(&body)
    }
}

// gpl licensed from wikipedia https://commons.wikimedia.org/wiki/File:Generic_Feed-icon.svg
pub const RSS_SVG: &str = include_str!("rss.url");
